from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas
import pendulum

from tecton._internals.run_api import validate_batch_mock_inputs_keys
from tecton.framework.data_frame import TectonDataFrame
from tecton.snowflake_context import SnowflakeContext
from tecton_core import conf
from tecton_core import schema_derivation_utils as core_schema_derivation_utils
from tecton_core import specs
from tecton_core.errors import TectonSnowflakeNotImplementedError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_core.materialization_context import BoundMaterializationContext
from tecton_core.time_utils import get_timezone_aware_datetime
from tecton_proto.args import feature_view_pb2 as feature_view__args_pb2
from tecton_proto.args import virtual_data_source_pb2 as virtual_data_source__args_pb2
from tecton_proto.common import schema_pb2
from tecton_proto.common import spark_schema_pb2
from tecton_snowflake import schema_derivation_utils
from tecton_snowflake import sql_helper


def get_historical_features(
    feature_set_config: FeatureSetConfig,
    spine: Optional[Union["snowflake.snowpark.DataFrame", pandas.DataFrame, TectonDataFrame, str]] = None,
    timestamp_key: Optional[str] = None,
    include_feature_view_timestamp_columns: bool = False,
    from_source: Optional[bool] = None,
    save: bool = False,
    save_as: Optional[str] = None,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    entities: Optional[Union["snowflake.snowpark.DataFrame", pandas.DataFrame, TectonDataFrame]] = None,
    append_prefix: bool = True,  # Whether to append the prefix to the feature column name
) -> TectonDataFrame:
    # TODO(TEC-6991): Dataset doesn't really work with snowflake as it has spark dependency.
    # Need to rewrite it with snowflake context or remove this param for snowflake.
    if save or save_as is not None:
        raise TectonSnowflakeNotImplementedError("save is not supported for Snowflake")

    start_time = get_timezone_aware_datetime(start_time)
    end_time = get_timezone_aware_datetime(end_time)

    if conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
        if entities is not None:
            # Convert entities to a snowflake dataframe
            if isinstance(entities, pandas.DataFrame):
                entities = TectonDataFrame._create(entities).to_snowflake()
            elif isinstance(entities, TectonDataFrame):
                entities = entities.to_snowflake()

        return TectonDataFrame._create_with_snowflake(
            sql_helper.get_historical_features_with_snowpark(
                spine=spine,
                session=SnowflakeContext.get_instance().get_session(),
                timestamp_key=timestamp_key,
                feature_set_config=feature_set_config,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                start_time=start_time,
                end_time=end_time,
                entities=entities,
                append_prefix=append_prefix,
                from_source=from_source,
            )
        )
    else:
        if timestamp_key is None and spine is not None:
            raise TectonSnowflakeNotImplementedError("timestamp_key must be specified using Snowflake without Snowpark")
        if entities is not None:
            raise TectonSnowflakeNotImplementedError("entities is only supported for Snowflake with Snowpark enabled")
        return TectonDataFrame._create(
            sql_helper.get_historical_features(
                spine=spine,
                connection=SnowflakeContext.get_instance().get_connection(),
                timestamp_key=timestamp_key,
                feature_set_config=feature_set_config,
                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                start_time=start_time,
                end_time=end_time,
                append_prefix=append_prefix,
                from_source=from_source,
            )
        )


def run_batch(
    fd: FeatureDefinition,
    mock_inputs: Dict[str, pandas.DataFrame],
    feature_start_time: datetime,
    feature_end_time: datetime,
    aggregation_level: Optional[str],
) -> TectonDataFrame:
    validate_batch_mock_inputs_keys(mock_inputs, fd)
    mock_sql_inputs = None

    feature_start_time = get_timezone_aware_datetime(feature_start_time)
    feature_end_time = get_timezone_aware_datetime(feature_end_time)

    if fd.is_temporal_aggregate:
        for feature in fd.fv_spec.aggregate_features:
            aggregate_function = sql_helper.AGGREGATION_PLANS[feature.function]
            if not aggregate_function:
                raise TectonSnowflakeNotImplementedError(
                    f"Unsupported aggregation function {feature.function} in snowflake pipeline"
                )

    session = None
    connection = None
    if conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
        session = SnowflakeContext.get_instance().get_session()
    else:
        connection = SnowflakeContext.get_instance().get_connection()

    if mock_inputs:
        mock_sql_inputs = {
            key: sql_helper.generate_sql_table_from_pandas_df(
                df=df, session=session, connection=connection, table_name=f"_TT_TEMP_INPUT_{key.upper()}_TABLE"
            )
            for (key, df) in mock_inputs.items()
        }

    # Validate input start and end times. Set defaults if None.
    sql_str = sql_helper.generate_run_batch_sql(
        feature_definition=fd,
        feature_start_time=feature_start_time,
        feature_end_time=feature_end_time,
        aggregation_level=aggregation_level,
        mock_sql_inputs=mock_sql_inputs,
        materialization_context=BoundMaterializationContext._create_internal(
            pendulum.instance(feature_start_time),
            pendulum.instance(feature_end_time),
            fd.fv_spec.batch_schedule,
        ),
        session=session,
        from_source=True,  # For run() we don't use materialized data
    )
    if session is not None:
        return TectonDataFrame._create_with_snowflake(session.sql(sql_str))
    else:
        cur = connection.cursor()
        cur.execute(sql_str, _statement_params={"SF_PARTNER": "tecton-ai"})
        return TectonDataFrame._create(cur.fetch_pandas_all())


def get_dataframe_for_data_source(
    data_source: specs.DataSourceSpec,
    start_time: datetime = None,
    end_time: datetime = None,
) -> TectonDataFrame:
    if not conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
        raise TectonSnowflakeNotImplementedError("get_dataframe is only supported with Snowpark enabled")

    start_time = get_timezone_aware_datetime(start_time)
    end_time = get_timezone_aware_datetime(end_time)

    session = SnowflakeContext.get_instance().get_session()
    return TectonDataFrame._create_with_snowflake(
        sql_helper.get_dataframe_for_data_source(session, data_source.batch_source, start_time, end_time)
    )


# For notebook driven development
def derive_batch_schema(
    ds_args: virtual_data_source__args_pb2.VirtualDataSourceArgs,
) -> spark_schema_pb2.SparkSchema:
    if not ds_args.HasField("snowflake_ds_config"):
        raise ValueError(f"Invalid batch source args: {ds_args}")

    connection = SnowflakeContext.get_instance().get_connection()
    return schema_derivation_utils.get_snowflake_schema(ds_args, connection)


def derive_view_schema_for_feature_view(
    feature_view_args: feature_view__args_pb2.FeatureViewArgs,
    transformation_specs: List[specs.TransformationSpec],
    data_source_specs: List[specs.DataSourceSpec],
) -> schema_pb2.Schema:
    connection = SnowflakeContext.get_instance().get_connection()
    return schema_derivation_utils.get_feature_view_view_schema(
        feature_view_args, transformation_specs, data_source_specs, connection
    )


def derive_materialization_schema_for_feature_view(
    view_schema: schema_pb2.Schema,
    feature_view_args: feature_view__args_pb2.FeatureViewArgs,
) -> schema_pb2.Schema:
    is_aggregate = len(feature_view_args.materialized_feature_view_args.aggregations) > 0
    if not is_aggregate:
        return view_schema

    materialization_schema = core_schema_derivation_utils.compute_aggregate_materialization_schema_from_view_schema(
        view_schema, feature_view_args, is_spark=False
    )

    # Make column name uppercase.
    for column in materialization_schema.columns:
        column.name = column.name.upper()
    return materialization_schema



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/serving_status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/data/serving_status.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1ctecton_proto/common/id.proto\"m\n\rServingStatus\x12\x44\n\rserving_state\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.data.ServingStateR\x0cservingState\x12\x16\n\x06\x65rrors\x18\x02 \x03(\tR\x06\x65rrors\"\xcd\x01\n\x0bStatusRange\x12\x43\n\x0f\x62\x65gin_inclusive\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x62\x65ginInclusive\x12?\n\rend_exclusive\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x65ndExclusive\x12\x38\n\x06status\x18\x03 \x01(\x0b\x32 .tecton_proto.data.ServingStatusR\x06status\"\xc5\x01\n FeatureViewMaterializationRanges\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12`\n\x1c\x62\x61tch_materialization_ranges\x18\x02 \x03(\x0b\x32\x1e.tecton_proto.data.StatusRangeR\x1a\x62\x61tchMaterializationRanges\"\xa0\x05\n\x14ServingStatusSummary\x12X\n\x18offline_readiness_ranges\x18\x01 \x03(\x0b\x32\x1e.tecton_proto.data.StatusRangeR\x16offlineReadinessRanges\x12V\n\x17online_readiness_ranges\x18\x07 \x03(\x0b\x32\x1e.tecton_proto.data.StatusRangeR\x15onlineReadinessRanges\x12u\n\x1c\x62\x61tch_materialization_ranges\x18\x04 \x03(\x0b\x32\x33.tecton_proto.data.FeatureViewMaterializationRangesR\x1a\x62\x61tchMaterializationRanges\x12L\n\x14streaming_start_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x12streamingStartTime\x12Z\n\x18streaming_serving_status\x18\x02 \x01(\x0b\x32 .tecton_proto.data.ServingStatusR\x16streamingServingStatus\x12O\n\x16most_recent_ready_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x13mostRecentReadyTime\x12\x64\n!most_recent_batch_processing_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x1dmostRecentBatchProcessingTime\"\xc7\x01\n\x1f\x46\x65\x61tureViewServingStatusSummary\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x03 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12]\n\x16serving_status_summary\x18\x02 \x01(\x0b\x32\'.tecton_proto.data.ServingStatusSummaryR\x14servingStatusSummaryJ\x04\x08\x01\x10\x02\"\xa6\x02\n FullFeatureServiceServingSummary\x12{\n&feature_service_serving_status_summary\x18\x02 \x01(\x0b\x32\'.tecton_proto.data.ServingStatusSummaryR\"featureServiceServingStatusSummary\x12\x84\x01\n%feature_view_serving_status_summaries\x18\x03 \x03(\x0b\x32\x32.tecton_proto.data.FeatureViewServingStatusSummaryR!featureViewServingStatusSummaries*\xcd\x01\n\x0cServingState\x12\x19\n\x15SERVING_STATE_UNKNOWN\x10\x00\x12\x1a\n\x16SERVING_STATE_DISABLED\x10\x01\x12!\n\x1dSERVING_STATE_NOT_ENOUGH_DATA\x10\x02\x12\x19\n\x15SERVING_STATE_PENDING\x10\x03\x12\x14\n\x10SERVING_STATE_OK\x10\x04\x12\x17\n\x13SERVING_STATE_ERROR\x10\x05\x12\x19\n\x15SERVING_STATE_RUNNING\x10\x06\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.serving_status_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _SERVINGSTATE._serialized_start=1818
  _SERVINGSTATE._serialized_end=2023
  _SERVINGSTATUS._serialized_start=124
  _SERVINGSTATUS._serialized_end=233
  _STATUSRANGE._serialized_start=236
  _STATUSRANGE._serialized_end=441
  _FEATUREVIEWMATERIALIZATIONRANGES._serialized_start=444
  _FEATUREVIEWMATERIALIZATIONRANGES._serialized_end=641
  _SERVINGSTATUSSUMMARY._serialized_start=644
  _SERVINGSTATUSSUMMARY._serialized_end=1316
  _FEATUREVIEWSERVINGSTATUSSUMMARY._serialized_start=1319
  _FEATUREVIEWSERVINGSTATUSSUMMARY._serialized_end=1518
  _FULLFEATURESERVICESERVINGSUMMARY._serialized_start=1521
  _FULLFEATURESERVICESERVINGSUMMARY._serialized_end=1815
# @@protoc_insertion_point(module_scope)



from typing import Optional
from typing import Tuple
from urllib.parse import urlparse


def get_kafka_secrets(
    ssl_keystore_location: str, ssl_keystore_password_secret_id: Optional[str] = None
) -> Tuple[str, Optional[str]]:
    local_keystore_password = None
    parsed_uri = urlparse(ssl_keystore_location, allow_fragments=False)

    if parsed_uri.scheme == "dbfs":
        path = parsed_uri.path.lstrip("/")
        local_keystore_loc = f"/dbfs/{path}"
    else:
        # This location must have been written to by a bootstrap script that copies the file from s3 to local.
        file_name = ssl_keystore_location.split("/")[-1]
        local_keystore_loc = f"/var/kafka-credentials/{file_name}"

    if ssl_keystore_password_secret_id:
        from tecton_core import conf

        local_keystore_password = conf.get_or_none(ssl_keystore_password_secret_id)

    return local_keystore_loc, local_keystore_password



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/acl.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1btecton_proto/auth/acl.proto\x12\x11tecton_proto.auth\x1a\x1ctecton_proto/common/id.proto\"^\n\x03\x41\x63l\x12\x1e\n\npermission\x18\x01 \x01(\tR\npermission\x12\x37\n\x0b\x61pi_key_ids\x18\x02 \x03(\x0b\x32\x17.tecton_proto.common.IdR\tapiKeyIdsB;\n\x0f\x63om.tecton.authP\x01Z&github.com/tecton-ai/tecton_proto/auth')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.acl_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authP\001Z&github.com/tecton-ai/tecton_proto/auth'
  _ACL._serialized_start=80
  _ACL._serialized_end=174
# @@protoc_insertion_point(module_scope)



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1680637774', 'BUILD_TIMESTAMP': '1680637774', 'GIT_COMMIT': 'daea5c48b146939ac640aee9282fc455e4041cfc', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-medium-76ddd7c467-jthkn', 'BUILD_USER': 'buildkite-agent'}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/onboarding.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"tecton_proto/data/onboarding.proto\x12\x11tecton_proto.data\"l\n\x10OnboardingStatus\x12\x17\n\x07user_id\x18\x01 \x01(\tR\x06userId\x12?\n\x06status\x18\x02 \x01(\x0e\x32\'.tecton_proto.data.OnboardingStatusEnumR\x06status\"\xdd\x01\n\x1b\x44\x61taPlatformSetupTaskStatus\x12*\n\x11task_display_name\x18\x01 \x01(\tR\x0ftaskDisplayName\x12S\n\x0btask_status\x18\x02 \x01(\x0e\x32\x32.tecton_proto.data.DataPlatformSetupTaskStatusEnumR\ntaskStatus\x12#\n\rerror_message\x18\x03 \x01(\tR\x0c\x65rrorMessage\x12\x18\n\x07\x64\x65tails\x18\x04 \x01(\tR\x07\x64\x65tails*x\n\x14OnboardingStatusEnum\x12\x1d\n\x19ONBOARDING_STATUS_UNKNOWN\x10\x00\x12 \n\x1cONBOARDING_STATUS_INCOMPLETE\x10\x01\x12\x1f\n\x1bONBOARDING_STATUS_COMPLETED\x10\x02*\xa3\x01\n\x1f\x44\x61taPlatformSetupTaskStatusEnum\x12\x17\n\x13TASK_STATUS_UNKNOWN\x10\x00\x12\x1b\n\x17TASK_STATUS_NOT_STARTED\x10\x01\x12\x17\n\x13TASK_STATUS_RUNNING\x10\x02\x12\x19\n\x15TASK_STATUS_SUCCEEDED\x10\x03\x12\x16\n\x12TASK_STATUS_FAILED\x10\x04\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.onboarding_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _ONBOARDINGSTATUSENUM._serialized_start=391
  _ONBOARDINGSTATUSENUM._serialized_end=511
  _DATAPLATFORMSETUPTASKSTATUSENUM._serialized_start=514
  _DATAPLATFORMSETUPTASKSTATUSENUM._serialized_end=677
  _ONBOARDINGSTATUS._serialized_start=57
  _ONBOARDINGSTATUS._serialized_end=165
  _DATAPLATFORMSETUPTASKSTATUS._serialized_start=168
  _DATAPLATFORMSETUPTASKSTATUS._serialized_end=389
# @@protoc_insertion_point(module_scope)



from typing import TYPE_CHECKING

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession

if TYPE_CHECKING:
    from tecton_proto.materialization.params_pb2 import MaterializationTaskParams


_active_metrics_collector = None


def get_active_metrics_collector():
    if _active_metrics_collector is None:
        return NoopMetricsCollector()

    return _active_metrics_collector


class MetricsCollector:
    def observe(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError

    def publish(self):
        raise NotImplementedError


class NoopMetricsCollector(MetricsCollector):
    def observe(self, df: DataFrame) -> DataFrame:
        return df

    def publish(self):
        pass


class SparkMetricsCollector(MetricsCollector):
    def __init__(self, jvm_collector):
        self._jvm_collector = jvm_collector

    def observe(self, df: DataFrame) -> DataFrame:
        new_jdf = self._jvm_collector.observe(df._jdf)
        return DataFrame(new_jdf, df.sql_ctx)

    def publish(self):
        self._jvm_collector.publish()


def create_feature_metrics_collector(spark: SparkSession, params: "MaterializationTaskParams") -> MetricsCollector:
    global _active_metrics_collector

    if not params.HasField("data_observability_config"):
        return NoopMetricsCollector()

    config = params.data_observability_config
    if not config.enabled:
        return NoopMetricsCollector()

    jvm_collector = spark._jvm.com.tecton.dataobs.spark.MetricsCollector.fromMaterializationTaskParams(
        params.SerializeToString(), spark._jsparkSession
    )
    _active_metrics_collector = SparkMetricsCollector(jvm_collector)
    return _active_metrics_collector



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/materialization/materialization_task.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.spark_common import libraries_pb2 as tecton__proto_dot_spark__common_dot_libraries__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.data import feature_view_pb2 as tecton__proto_dot_data_dot_feature__view__pb2
from tecton_proto.args import feature_view_pb2 as tecton__proto_dot_args_dot_feature__view__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2
from tecton_proto.spark_api import jobs_pb2 as tecton__proto_dot_spark__api_dot_jobs__pb2
from tecton_proto.common import fco_locator_pb2 as tecton__proto_dot_common_dot_fco__locator__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.materialization import materialization_states_pb2 as tecton__proto_dot_materialization_dot_materialization__states__pb2
from tecton_proto.materialization import spark_cluster_pb2 as tecton__proto_dot_materialization_dot_spark__cluster__pb2
from tecton_proto.snowflake import location_pb2 as tecton__proto_dot_snowflake_dot_location__pb2
from tecton_proto.online_store_writer import config_pb2 as tecton__proto_dot_online__store__writer_dot_config__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7tecton_proto/materialization/materialization_task.proto\x12\x1ctecton_proto.materialization\x1a)tecton_proto/spark_common/libraries.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a$tecton_proto/data/feature_view.proto\x1a$tecton_proto/args/feature_view.proto\x1a$tecton_proto/args/diff_options.proto\x1a!tecton_proto/spark_api/jobs.proto\x1a%tecton_proto/common/fco_locator.proto\x1a\x1ctecton_proto/common/id.proto\x1a\x39tecton_proto/materialization/materialization_states.proto\x1a\x30tecton_proto/materialization/spark_cluster.proto\x1a%tecton_proto/snowflake/location.proto\x1a-tecton_proto/online_store_writer/config.proto\"\xb5\x01\n\"MaterializationTaskStateTransition\x12U\n\ntask_state\x18\x01 \x01(\x0e\x32\x36.tecton_proto.materialization.MaterializationTaskStateR\ttaskState\x12\x38\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\"\xdb\x04\n\x1e\x42\x61tchMaterializationParameters\x12\x46\n\x11window_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0fwindowStartTime\x12\x42\n\x0fwindow_end_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rwindowEndTime\x12H\n\x12\x66\x65\x61ture_start_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12\x1d\n\ntile_count\x18\x05 \x01(\x05R\ttileCount\x12@\n\x1dwrite_to_online_feature_store\x18\x08 \x01(\x08R\x19writeToOnlineFeatureStore\x12\x42\n\x1ewrite_to_offline_feature_store\x18\x06 \x01(\x08R\x1awriteToOfflineFeatureStore\x12\x32\n\x15is_overwrite_backfill\x18\t \x01(\x08R\x13isOverwriteBackfill\x12\x32\n\x15is_bootstrap_backfill\x18\n \x01(\x08R\x13isBootstrapBackfillJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x07\x10\x08\"\xc8\x01\n\x1fIngestMaterializationParameters\x12@\n\x1dwrite_to_online_feature_store\x18\x08 \x01(\x08R\x19writeToOnlineFeatureStore\x12\x42\n\x1ewrite_to_offline_feature_store\x18\x06 \x01(\x08R\x1awriteToOfflineFeatureStore\x12\x1f\n\x0bingest_path\x18\r \x01(\tR\ningestPath\"\xae\x01\n\x12\x44\x65letionParameters\x12\x31\n\x15online_join_keys_path\x18\x01 \x01(\tR\x12onlineJoinKeysPath\x12\x33\n\x16offline_join_keys_path\x18\x02 \x01(\tR\x13offlineJoinKeysPath\x12\x16\n\x06online\x18\x03 \x01(\x08R\x06online\x12\x18\n\x07offline\x18\x04 \x01(\x08R\x07offline\"\x84\x01\n\x1a\x44\x65ltaMaintenanceParameters\x12+\n\x11generate_manifest\x18\x01 \x01(\x08R\x10generateManifest\x12\x39\n\nperiod_end\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tperiodEnd\"-\n\x1fStreamMaterializationParametersJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03\"\xd0\x0c\n\x13MaterializationTask\x12O\n\x17materialization_task_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x15materializationTaskId\x12X\n\x17id_feature_view_locator\x18\x10 \x01(\x0b\x32!.tecton_proto.common.IdFcoLocatorR\x14idFeatureViewLocator\x12\x44\n\x1ematerialization_serial_version\x18\x0e \x01(\x05R\x1cmaterializationSerialVersion\x12i\n\x10\x62\x61tch_parameters\x18\x03 \x01(\x0b\x32<.tecton_proto.materialization.BatchMaterializationParametersH\x00R\x0f\x62\x61tchParameters\x12l\n\x11stream_parameters\x18\x04 \x01(\x0b\x32=.tecton_proto.materialization.StreamMaterializationParametersH\x00R\x10streamParameters\x12l\n\x11ingest_parameters\x18\x11 \x01(\x0b\x32=.tecton_proto.materialization.IngestMaterializationParametersH\x00R\x10ingestParameters\x12\x63\n\x13\x64\x65letion_parameters\x18\x13 \x01(\x0b\x32\x30.tecton_proto.materialization.DeletionParametersH\x00R\x12\x64\x65letionParameters\x12|\n\x1c\x64\x65lta_maintenance_parameters\x18\x18 \x01(\x0b\x32\x38.tecton_proto.materialization.DeltaMaintenanceParametersH\x00R\x1a\x64\x65ltaMaintenanceParameters\x12I\n!spark_cluster_environment_version\x18\r \x01(\x05R\x1esparkClusterEnvironmentVersion\x12m\n\x11state_transitions\x18\x08 \x03(\x0b\x32@.tecton_proto.materialization.MaterializationTaskStateTransitionR\x10stateTransitions\x12Q\n&number_of_attempts_from_old_executions\x18\x12 \x01(\x05R!numberOfAttemptsFromOldExecutions\x12#\n\rerror_message\x18\t \x01(\tR\x0c\x65rrorMessage\x12\x39\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x12\x41\n\x08\x61ttempts\x18\x14 \x03(\x0b\x32%.tecton_proto.materialization.AttemptR\x08\x61ttempts\x12_\n\x0b\x65nrichments\x18\xe8\x07 \x01(\x0b\x32<.tecton_proto.materialization.MaterializationTaskEnrichmentsR\x0b\x65nrichments\x12O\n\rcanary_params\x18\x15 \x01(\x0b\x32*.tecton_proto.materialization.CanaryParamsR\x0c\x63\x61naryParams\x12-\n\x12manually_triggered\x18\x16 \x01(\x08R\x11manuallyTriggered\x12\'\n\x0fmanaged_retries\x18\x17 \x01(\x08R\x0emanagedRetriesB\x0b\n\ttask_typeJ\x04\x08\x02\x10\x03J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\n\x10\x0bJ\x04\x08\x0f\x10\x10\"\xee\x03\n\x0c\x43\x61naryParams\x12\x1b\n\tcanary_id\x18\x01 \x01(\tR\x08\x63\x61naryId\x12r\n\x14\x63\x61nary_online_config\x18\x02 \x01(\x0b\x32@.tecton_proto.online_store_writer.OnlineStoreWriterConfigurationR\x12\x63\x61naryOnlineConfig\x12G\n canary_streaming_checkpoint_path\x18\x03 \x01(\tR\x1d\x63\x61naryStreamingCheckpointPath\x12\\\n\x18\x63\x61nary_library_overrides\x18\x04 \x03(\x0b\x32\".tecton_proto.spark_common.LibraryR\x16\x63\x61naryLibraryOverrides\x12~\n canary_spark_cluster_environment\x18\x05 \x01(\x0b\x32\x35.tecton_proto.materialization.SparkClusterEnvironmentR\x1d\x63\x61narySparkClusterEnvironment\x12&\n\x0f\x63\x61nary_run_name\x18\x06 \x01(\tR\rcanaryRunName\"\x99\x01\n\x07\x41ttempt\x12\x38\n\x0bworkflow_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\nworkflowId\x12T\n\'has_legacy_materialization_task_attempt\x18\x02 \x01(\x08R#hasLegacyMaterializationTaskAttempt\"\xcb\x02\n)MaterializationTaskAttemptStateTransition\x12\x62\n\rattempt_state\x18\x01 \x01(\x0e\x32=.tecton_proto.materialization.MaterializationTaskAttemptStateR\x0c\x61ttemptState\x12#\n\rstate_message\x18\x03 \x01(\tR\x0cstateMessage\x12[\n\x12termination_reason\x18\x04 \x01(\x0e\x32,.tecton_proto.spark_api.RunTerminationReasonR\x11terminationReason\x12\x38\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\"\xc3\x08\n\x1aMaterializationTaskAttempt\x12^\n\x1fmaterialization_task_attempt_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x1cmaterializationTaskAttemptId\x12O\n\x17materialization_task_id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x15materializationTaskId\x12g\n\x15\x65xecution_environment\x18\x0e \x01(\x0e\x32\x32.tecton_proto.materialization.ExecutionEnvironmentR\x14\x65xecutionEnvironment\x12\x15\n\x06run_id\x18\r \x01(\tR\x05runId\x12 \n\x0crun_page_url\x18\x0c \x01(\tR\nrunPageUrl\x12t\n\x11state_transitions\x18\x04 \x03(\x0b\x32G.tecton_proto.materialization.MaterializationTaskAttemptStateTransitionR\x10stateTransitions\x12%\n\x0e\x61ttempt_number\x18\x0b \x01(\x05R\rattemptNumber\x12\x39\n\ncreated_at\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x12G\n\x0e\x63luster_config\x18\x0f \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\rclusterConfig\x12\x8f\x01\n\x19online_store_copier_tasks\x18\x11 \x03(\x0b\x32T.tecton_proto.materialization.MaterializationTaskAttempt.OnlineStoreCopierTasksEntryR\x16onlineStoreCopierTasks\x12R\n\x0esnowflake_data\x18\x12 \x01(\x0b\x32+.tecton_proto.materialization.SnowflakeDataR\rsnowflakeData\x1a~\n\x1bOnlineStoreCopierTasksEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12I\n\x05value\x18\x02 \x01(\x0b\x32\x33.tecton_proto.materialization.OnlineStoreCopierTaskR\x05value:\x02\x38\x01J\x04\x08\x03\x10\x04J\x04\x08\x05\x10\tJ\x04\x08\x10\x10\x11\"\xe2\x01\n\x1eMaterializationTaskEnrichments\x12q\n\x19spark_cluster_environment\x18\x03 \x01(\x0b\x32\x35.tecton_proto.materialization.SparkClusterEnvironmentR\x17sparkClusterEnvironment\x12\x41\n\x0c\x66\x65\x61ture_view\x18\x04 \x01(\x0b\x32\x1e.tecton_proto.data.FeatureViewR\x0b\x66\x65\x61tureViewJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03\"\xd5\x01\n$OnlineStoreCopierTaskStateTransition\x12N\n\x05state\x18\x01 \x01(\x0e\x32\x38.tecton_proto.materialization.OnlineStoreCopierTaskStateR\x05state\x12#\n\rstate_message\x18\x03 \x01(\tR\x0cstateMessage\x12\x38\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\"\xc7\x02\n\x15OnlineStoreCopierTask\x12V\n\x08\x61ttempts\x18\x02 \x03(\x0b\x32:.tecton_proto.materialization.OnlineStoreCopierTaskAttemptR\x08\x61ttempts\x12\x1d\n\ntotal_rows\x18\x04 \x01(\x03R\ttotalRows\x12%\n\x0e\x63ompleted_rows\x18\x03 \x01(\x03R\rcompletedRows\x12\x37\n\x18last_progress_request_id\x18\x05 \x01(\tR\x15lastProgressRequestId\x12Q\n\x17last_progress_send_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x14lastProgressSendTimeJ\x04\x08\x01\x10\x02\"\x8f\x01\n\x1cOnlineStoreCopierTaskAttempt\x12o\n\x11state_transitions\x18\x01 \x03(\x0b\x32\x42.tecton_proto.materialization.OnlineStoreCopierTaskStateTransitionR\x10stateTransitions\"\x80\x01\n\rSnowflakeData\x12J\n\rbase_location\x18\x02 \x01(\x0b\x32%.tecton_proto.snowflake.StageLocationR\x0c\x62\x61seLocation\x12\x1d\n\ntotal_rows\x18\x03 \x01(\x03R\ttotalRowsJ\x04\x08\x01\x10\x02\x42\x1e\n\x1a\x63om.tecton.materializationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.materialization.materialization_task_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.materializationP\001'
  _MATERIALIZATIONTASKATTEMPT_ONLINESTORECOPIERTASKSENTRY._options = None
  _MATERIALIZATIONTASKATTEMPT_ONLINESTORECOPIERTASKSENTRY._serialized_options = b'8\001'
  _MATERIALIZATIONTASKSTATETRANSITION._serialized_start=579
  _MATERIALIZATIONTASKSTATETRANSITION._serialized_end=760
  _BATCHMATERIALIZATIONPARAMETERS._serialized_start=763
  _BATCHMATERIALIZATIONPARAMETERS._serialized_end=1366
  _INGESTMATERIALIZATIONPARAMETERS._serialized_start=1369
  _INGESTMATERIALIZATIONPARAMETERS._serialized_end=1569
  _DELETIONPARAMETERS._serialized_start=1572
  _DELETIONPARAMETERS._serialized_end=1746
  _DELTAMAINTENANCEPARAMETERS._serialized_start=1749
  _DELTAMAINTENANCEPARAMETERS._serialized_end=1881
  _STREAMMATERIALIZATIONPARAMETERS._serialized_start=1883
  _STREAMMATERIALIZATIONPARAMETERS._serialized_end=1928
  _MATERIALIZATIONTASK._serialized_start=1931
  _MATERIALIZATIONTASK._serialized_end=3547
  _CANARYPARAMS._serialized_start=3550
  _CANARYPARAMS._serialized_end=4044
  _ATTEMPT._serialized_start=4047
  _ATTEMPT._serialized_end=4200
  _MATERIALIZATIONTASKATTEMPTSTATETRANSITION._serialized_start=4203
  _MATERIALIZATIONTASKATTEMPTSTATETRANSITION._serialized_end=4534
  _MATERIALIZATIONTASKATTEMPT._serialized_start=4537
  _MATERIALIZATIONTASKATTEMPT._serialized_end=5628
  _MATERIALIZATIONTASKATTEMPT_ONLINESTORECOPIERTASKSENTRY._serialized_start=5484
  _MATERIALIZATIONTASKATTEMPT_ONLINESTORECOPIERTASKSENTRY._serialized_end=5610
  _MATERIALIZATIONTASKENRICHMENTS._serialized_start=5631
  _MATERIALIZATIONTASKENRICHMENTS._serialized_end=5857
  _ONLINESTORECOPIERTASKSTATETRANSITION._serialized_start=5860
  _ONLINESTORECOPIERTASKSTATETRANSITION._serialized_end=6073
  _ONLINESTORECOPIERTASK._serialized_start=6076
  _ONLINESTORECOPIERTASK._serialized_end=6403
  _ONLINESTORECOPIERTASKATTEMPT._serialized_start=6406
  _ONLINESTORECOPIERTASKATTEMPT._serialized_end=6549
  _SNOWFLAKEDATA._serialized_start=6552
  _SNOWFLAKEDATA._serialized_end=6680
# @@protoc_insertion_point(module_scope)



from typing import Dict
from typing import List

import pandas

from tecton_athena import pipeline_helper
from tecton_core import feature_set_config
from tecton_core import logger as logger_lib
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_proto.args.transformation_pb2 import TransformationMode

logger = logger_lib.get_logger("AthenaConnector")


def get_odfvs_from_feature_set_config(feature_set_config: FeatureSetConfig):
    feature_set_items = feature_set_config.definitions_and_configs
    odfvs = [fd.feature_definition for fd in feature_set_items if fd.feature_definition.is_on_demand]
    return odfvs


def _extract_inputs_for_odfv_from_data(
    data_df: pandas.DataFrame, odfv: FeatureDefinition
) -> Dict[str, pandas.DataFrame]:
    odfv_invocation_inputs = {}

    odfv_transformation_node = odfv.pipeline.root.transformation_node

    for input in odfv_transformation_node.inputs:
        input_name = input.arg_name
        input_df = None

        if input.node.HasField("request_data_source_node"):
            request_context_schema = input.node.request_data_source_node.request_context.schema
            request_context_fields = [f.name for f in request_context_schema.fields]

            for f in request_context_fields:
                if f not in data_df.columns:
                    raise TectonValidationError(
                        f"ODFV {odfv.name} has a dependency on the Request Data Source named '{input_name}'. Field {f} of this Request Data Source is not found in the spine. Available columns: {list(data_df.columns)}"
                    )

            input_df = data_df[request_context_fields]
        elif input.node.HasField("feature_view_node"):
            fv_features = feature_set_config.find_dependent_feature_set_items(
                odfv.fco_container, input.node, {}, odfv.id
            )[0]
            select_columns_and_rename_map = {}
            for f in fv_features.features:
                column_name = f"{fv_features.namespace}__{f}"
                mapped_name = f
                select_columns_and_rename_map[column_name] = mapped_name

            for f in select_columns_and_rename_map.keys():
                if f not in data_df.columns:
                    raise TectonValidationError(
                        f"ODFV {odfv.name} has a dependency on the Feature View '{input_name}'. Feature {f} of this Feature View is not found in the retrieved historical data. Available columns: {list(data_df.columns)}"
                    )
            # Let's select all of the features of the input FV from data_df
            input_df = data_df.rename(columns=select_columns_and_rename_map)[[*select_columns_and_rename_map.values()]]
        else:
            raise Exception(f"Unexpected input found ({input_name}) on ODFV {odfv.name}")

        odfv_invocation_inputs[input_name] = input_df

    return odfv_invocation_inputs


def _run_odfv(data_df: pandas.DataFrame, odfv: FeatureDefinition) -> pandas.DataFrame:
    transformation_mode = odfv.transformations[0].transformation_mode

    odfv_pipeline = pipeline_helper.build_odfv_execution_pipeline(
        pipeline=odfv.pipeline, transformations=odfv.transformations, name=odfv.name
    )

    if transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
        odfv_inputs = _extract_inputs_for_odfv_from_data(data_df, odfv)
        odfv_result_df = odfv_pipeline.execute_with_inputs(odfv_inputs)
        return odfv_result_df
    elif transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
        odfv_inputs = _extract_inputs_for_odfv_from_data(data_df, odfv)

        # The inputs are currently a mapping of input_name to pandas DF
        # We need turn the ODFV inputs from a pandas DF to a list of dictionaries
        # Then we need to iterate through all rows of the input data set, pass the input dicts into the ODFV
        # And finally convert the resulting list of dicts into a pandas DF
        for input_name in odfv_inputs.keys():
            # Map pandas DFs to List of dicts (one dict per row)
            odfv_inputs[input_name] = odfv_inputs[input_name].to_dict("records")

        odfv_result_list = []

        num_rows = len(data_df)
        if num_rows > 100:
            logger.warn(
                f"Executing ODFV {odfv.name} for {len(data_df)} rows. The ODFV will be executed row by row and may take a while to complete..."
            )

        for row_index in range(num_rows):
            # Iterate through all rows of the data and invoke the ODFV
            row_odfv_inputs = {}
            for input_name in odfv_inputs.keys():
                row_odfv_inputs[input_name] = odfv_inputs[input_name][row_index]

            odfv_result_dict = odfv_pipeline.execute_with_inputs(row_odfv_inputs)
            odfv_result_list.append(odfv_result_dict)
        return pandas.DataFrame.from_dict(odfv_result_list)
    else:
        raise TectonValidationError(f"ODFV {odfv.name} has an unexpected transformation mode: {transformation_mode}")


def run_and_append_on_demand_features_to_historical_data(data_df: pandas.DataFrame, odfvs: List[FeatureDefinition]):
    for odfv in odfvs:
        odfv_result_df = _run_odfv(data_df, odfv)
        # We're performing an index based merge between the historical data and the ODFV results
        # That's safe to do because the ODFV function isn't expected to change the order of rows
        data_df = data_df.merge(odfv_result_df, left_index=True, right_index=True)
    return data_df



from typing import Dict
from typing import List
from typing import Optional

import attrs
import pendulum
import pyspark
from pyspark.sql import functions
from pyspark.sql.functions import expr

from tecton_core.query_consts import ANCHOR_TIME
from tecton_spark.query.node import SparkExecNode
from tecton_spark.time_utils import convert_epoch_to_timestamp_column
from tecton_spark.time_utils import convert_timestamp_to_epoch


@attrs.frozen
class AddAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_format_version: int
    batch_schedule: int
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(
            functions.col(self.timestamp_field), self.feature_store_format_version
        )
        df = input_df.withColumn(
            ANCHOR_TIME,
            anchor_time_val - anchor_time_val % self.batch_schedule,
        )
        return df


@attrs.frozen
class AddRetrievalAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    name: str
    feature_store_format_version: int
    batch_schedule: int
    tile_interval: int
    timestamp_field: str
    is_stream: bool
    data_delay_seconds: Optional[int] = 0

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(
            functions.col(self.timestamp_field) - expr(f"interval {self.data_delay_seconds} seconds"),
            self.feature_store_format_version,
        )
        # tile_interval will be 0 for continuous
        if self.tile_interval == 0:
            df = input_df.withColumn(ANCHOR_TIME, anchor_time_val)
        else:
            # For stream, we use the tile interval for bucketing since the data is available as soon as
            # the aggregation interval ends.
            # For BAFV, we use the batch schedule to get the last tile written.
            if self.is_stream:
                df = input_df.withColumn(
                    ANCHOR_TIME,
                    anchor_time_val - anchor_time_val % self.tile_interval - self.tile_interval,
                )
            else:
                df = input_df.withColumn(
                    ANCHOR_TIME,
                    anchor_time_val - anchor_time_val % self.batch_schedule - self.tile_interval,
                )
        return df


@attrs.frozen
class RenameColsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    mapping: Optional[Dict[str, str]]
    drop: Optional[List[str]]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        if self.mapping:
            for old_name, new_name in self.mapping.items():
                if new_name:
                    input_df = input_df.withColumnRenamed(old_name, new_name)
        if self.drop:
            for col in self.drop:
                input_df = input_df.drop(col)
        return input_df


@attrs.frozen
class ConvertEpochToTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_formats: Dict[str, int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        for name, feature_store_format_version in self.feature_store_formats.items():
            input_df = input_df.withColumn(
                name,
                convert_epoch_to_timestamp_column(functions.col(name), feature_store_format_version),
            )
        return input_df


@attrs.frozen
class AddEffectiveTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    effective_timestamp_name: str
    batch_schedule_seconds: int
    is_stream: bool
    data_delay_seconds: int
    is_temporal_aggregate: bool

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        # batch_schedule = 0 implies feature table.
        if self.batch_schedule_seconds == 0 or self.is_stream:
            effective_timestamp = functions.col(self.timestamp_field)
        else:
            slide_str = f"{self.batch_schedule_seconds} seconds"
            timestamp_col = functions.col(self.timestamp_field)
            # Timestamp of temporal aggregate is end of the anchor time window. Subtract 1 micro
            # to get the correct bucket for batch schedule.
            if self.is_temporal_aggregate:
                timestamp_col -= expr(f"interval 1 microseconds")
            window_spec = functions.window(timestamp_col, slide_str, slide_str)
            effective_timestamp = window_spec.end + expr(f"interval {self.data_delay_seconds} seconds")

        df = input_df.withColumn(self.effective_timestamp_name, effective_timestamp)
        return df


@attrs.frozen
class AddDurationSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    duration: pendulum.Duration
    new_column_name: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        df = input_df.withColumn(
            self.new_column_name,
            functions.col(self.timestamp_field) + expr(f"interval {self.duration.total_seconds()} seconds"),
        )
        return df


@attrs.frozen
class SelectDistinctSparkNode(SparkExecNode):
    input_node: SparkExecNode
    columns: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.select(self.columns).distinct()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/spark_schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/common/spark_schema.proto\x12\x13tecton_proto.common\"K\n\nSparkField\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12)\n\x10structfield_json\x18\x02 \x01(\tR\x0fstructfieldJson\"F\n\x0bSparkSchema\x12\x37\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x1f.tecton_proto.common.SparkFieldR\x06\x66ieldsB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.spark_schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _SPARKFIELD._serialized_start=63
  _SPARKFIELD._serialized_end=138
  _SPARKSCHEMA._serialized_start=140
  _SPARKSCHEMA._serialized_end=210
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import column_type_pb2 as tecton__proto_dot_common_dot_column__type__pb2
from tecton_proto.common import data_type_pb2 as tecton__proto_dot_common_dot_data__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n tecton_proto/common/schema.proto\x12\x13tecton_proto.common\x1a%tecton_proto/common/column_type.proto\x1a#tecton_proto/common/data_type.proto\"\xe4\x02\n\x06\x43olumn\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12$\n\x0eraw_spark_type\x18\x02 \x01(\tR\x0crawSparkType\x12,\n\x12raw_snowflake_type\x18\x04 \x01(\tR\x10rawSnowflakeType\x12O\n\x13\x66\x65\x61ture_server_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.common.ColumnTypeR\x11\x66\x65\x61tureServerType\x12I\n\x11offline_data_type\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x0fofflineDataType\x12V\n\x18\x66\x65\x61ture_server_data_type\x18\x06 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x15\x66\x65\x61tureServerDataType\"E\n\x06Schema\x12\x35\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x07\x63olumnsJ\x04\x08\x01\x10\x02\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _COLUMN._serialized_start=134
  _COLUMN._serialized_end=490
  _SCHEMA._serialized_start=492
  _SCHEMA._serialized_end=561
# @@protoc_insertion_point(module_scope)



import setuptools
import os

with open('README.md') as readme_f:
  long_description = readme_f.read()

packages = setuptools.find_packages()
for root, _, files in os.walk("tecton_proto"):
  if any([f.endswith("_pb2.py") for f in files]):
    packages.append(root.replace("/", "."))
for root, _, files in os.walk("protoc_gen_swagger"):
  if any([f.endswith("_pb2.py") for f in files]):
    packages.append(root.replace("/", "."))

setuptools.setup(
    classifiers=['Programming Language :: Python :: 3', 'Operating System :: OS Independent', 'License :: Other/Proprietary License'],
    python_requires='>=3.7.*',
    author='Tecton, Inc.',
    author_email='support@tecton.ai',
    url='https://tecton.ai',
    license='Tecton Proprietary',
    include_package_data=True,
    description='Tecton Python SDK',
    entry_points={'console_scripts': ['tecton=tecton.cli.cli:main'], 'pytest11': ['pytest_tecton=tecton.pytest_tecton']},
    name='tecton',
    version='0.7.0b10',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['attrs>=21.3.0', 'boto3', 'googleapis-common-protos~=1.52', 'jinja2~=3.0.3', 'numpy~=1.16', 'pathspec', 'pendulum~=2.1', 'protobuf~=3.20.0', 'pypika~=0.48.9', 'pytimeparse', 'pandas~=1.0', 'texttable', 'requests', 'colorama~=0.4', 'tqdm~=4.41', 'yaspin<3,>=0.16', 'typing-extensions~=4.1', 'pygments>=2.7.4', 'pytest', 'click~=8.0', 'typeguard~=2.0', 'sqlparse', 'semantic_version'],
    extras_require={'databricks-connect': ['databricks-connect[sql]~=9.1.23'], 'databricks-connect9': ['databricks-connect[sql]~=9.1.23'], 'databricks-connect10': ['databricks-connect[sql]~=10.4.12'], 'pyspark': ['pyspark[sql]~=3.1.2'], 'pyspark3': ['pyspark[sql]~=3.1.2'], 'pyspark3.1': ['pyspark[sql]~=3.1.2'], 'pyspark3.2': ['pyspark[sql]~=3.2.1'], 'snowflake': ['snowflake-connector-python[pandas]~=2.8'], 'snowpark': ['snowflake-snowpark-python[pandas]~=1.0.0'], 'athena': ['awswrangler~=2.15']},
    packages=packages,
)



import json
import logging
import os
import sys
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

from tecton_core import errors

logging.getLogger("boto3").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)

_ConfigSettings = Dict[str, Any]

_CONFIG_OVERRIDES: _ConfigSettings = {}


# Tecton conf unfortunately defines a top-level "set" method which shadows the "set" data type. Before defining the
# set method, get a reference to the set data type.
_set_type = set


class ConfSource(Enum):
    # (Always supported) This key can be overriden in the Python runtime.
    SESSION_OVERRIDE = 1
    # (Always supported) This key can be read from the local environment.
    OS_ENV = 2
    # This key can be written to and read from the tecton config file.
    LOCAL_TECTON_CONFIG = 3
    # This key can be written to and read from the tecton tokens file.
    LOCAL_TECTON_TOKENS = 4
    # This key can read from the MDS get-configs endpoint.
    REMOTE_MDS_CONFIG = 5
    # This key can read from the Databricks secrets manager.
    DATABRICKS_SECRET = 6
    # This key can read from the AWS secrets manager.
    AWS_SECRET_MANAGER = 7
    # The default value of this key was used. Used for debug printing. Not an "allowable" source.
    DEFAULT = 8
    # A value for this key was not found. Used for debug printing. Not an "allowable" source.
    NOT_FOUND = 9


RUNTIME_ALLOWED_SOURCES = (ConfSource.SESSION_OVERRIDE, ConfSource.OS_ENV)
DEFAULT_ALLOWED_SOURCES = (
    ConfSource.SESSION_OVERRIDE,
    ConfSource.OS_ENV,
    ConfSource.REMOTE_MDS_CONFIG,
    ConfSource.DATABRICKS_SECRET,
    ConfSource.AWS_SECRET_MANAGER,
)


class _Debugger(object):
    @classmethod
    def _debug_enabled(cls) -> bool:
        debug_value = _get_runtime_only("TECTON_DEBUG")
        return debug_value is not None and debug_value.lower() in ("1", "true", "yes")

    @classmethod
    def preamble(cls, key: str):
        if cls._debug_enabled():
            print(f"Looking up {key}", file=sys.stderr)

    @classmethod
    def print(cls, src: ConfSource, key: str, val: Optional[str] = None, details: Optional[str] = None):
        if not cls._debug_enabled():
            return

        if src == ConfSource.NOT_FOUND:
            print(f"Unable to find {key}\n", file=sys.stderr)
            return

        details_str = f"({details})" if details else ""
        val_str = val if val is not None else "not found"
        symbol_str = "[x]" if val is not None else "[ ]"
        print(symbol_str, f"{key} in {src.name} -> {val_str}", details_str, file=sys.stderr)


def set(key, value):
    _set(key, value)


def unset(key):
    del _CONFIG_OVERRIDES[key]


def _set(key, value):
    _CONFIG_OVERRIDES[key] = value


def _does_key_have_valid_prefix(key) -> bool:
    for prefix in _VALID_KEY_PREFIXES:
        if key.startswith(prefix):
            return True
    return False


def _get(key) -> Optional[str]:
    """Get the config value for the given key, or return None if not found."""
    _Debugger.preamble(key)

    if key in _VALID_KEYS_TO_ALLOWED_SOURCES:
        allowed_sources = _VALID_KEYS_TO_ALLOWED_SOURCES[key]
    elif _does_key_have_valid_prefix(key):
        allowed_sources = DEFAULT_ALLOWED_SOURCES
    else:
        raise errors.TectonInternalError(f"Tried accessing invalid configuration key '{key}'")

    # Session-scoped override.
    if ConfSource.SESSION_OVERRIDE in allowed_sources:
        val = _CONFIG_OVERRIDES.get(key)
        _Debugger.print(ConfSource.SESSION_OVERRIDE, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # Environment variable.
    if ConfSource.OS_ENV in allowed_sources:
        val = os.environ.get(key)
        _Debugger.print(ConfSource.OS_ENV, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # ~/.tecton/config
    if ConfSource.LOCAL_TECTON_CONFIG in allowed_sources:
        val = _LOCAL_TECTON_CONFIG.get(key)
        _Debugger.print(ConfSource.LOCAL_TECTON_CONFIG, key, val, details=str(_LOCAL_TECTON_CONFIG_FILE))
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # ~/.tecton/config.tokens
    if ConfSource.LOCAL_TECTON_TOKENS in allowed_sources:
        val = _LOCAL_TECTON_TOKENS.get(key)
        _Debugger.print(ConfSource.LOCAL_TECTON_TOKENS, key, val, details=str(_LOCAL_TECTON_CONFIG_FILE))
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    # Config from MDS
    if ConfSource.REMOTE_MDS_CONFIG in allowed_sources:
        val = _REMOTE_MDS_CONFIGS.get(key)
        _Debugger.print(ConfSource.REMOTE_MDS_CONFIG, key, val)
        if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
            return val

    if _get_runtime_env() == TectonEnv.UNKNOWN:
        # Fallback attempt to set env if user has not set it.
        _set_tecton_runtime_env()

    if ConfSource.DATABRICKS_SECRET in allowed_sources and not _get_runtime_only("TECTON_CONF_DISABLE_DBUTILS"):
        # Databricks secrets
        for scope in _get_secret_scopes():
            value = _get_from_db_secrets(key, scope)
            _Debugger.print(ConfSource.DATABRICKS_SECRET, key, value, details=f"{scope}:{key}")
            if value is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
                return value

    if ConfSource.AWS_SECRET_MANAGER in allowed_sources and not _get_runtime_only("TECTON_CONF_DISABLE_AWS_SECRETS"):
        # AWS secret manager
        for scope in _get_secret_scopes():
            value = _get_from_secretsmanager(key, scope)
            _Debugger.print(ConfSource.AWS_SECRET_MANAGER, key, value, details=f"{scope}/{key}")
            if value is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
                return value

    if key in _DEFAULTS:
        value = _DEFAULTS[key]()
        _Debugger.print(ConfSource.DEFAULT, key, value)
        return value

    _Debugger.print(ConfSource.NOT_FOUND, key)
    return None


def _get_runtime_only(key) -> Optional[str]:
    """An alternate _get() that will look up only from runtime sources. Used to avoid infinite loops."""
    if key not in _VALID_KEYS_TO_ALLOWED_SOURCES:
        raise errors.TectonInternalError(f"_get_runtime_only should only used with valid keys. {key}")

    allowed_sources = _VALID_KEYS_TO_ALLOWED_SOURCES[key]
    if _set_type(allowed_sources) != _set_type(RUNTIME_ALLOWED_SOURCES):
        raise errors.TectonInternalError(
            f"_get_runtime_only should only used with keys that only allow run time sources."
        )

    # Session-scoped override.
    val = _CONFIG_OVERRIDES.get(key)
    if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
        return val

    # Environment variable.
    val = os.environ.get(key)
    if val is not None:  # NOTE: check explicitly against None so we can set a value to False or ""
        return val

    if key in _DEFAULTS:
        value = _DEFAULTS[key]()
        return value

    return None


def get_or_none(key) -> Optional[str]:
    return _get(key)


def get_or_raise(key) -> str:
    val = _get(key)
    if val is None:
        raise errors.TectonInternalError(f"{key} not set")
    return val


def get_bool(key) -> bool:
    val = _get(key)
    if val is None:
        return False
    # bit of a hack for if people set a boolean value in a local override
    if isinstance(val, bool):
        return val
    if not isinstance(val, str):
        raise ValueError(f"{key} should be an instance of str, not {type(val)}")
    if val.lower() in {"yes", "true"}:
        return True
    if val.lower() in {"no", "false"}:
        return False
    raise ValueError(f"{key} should be 'true' or 'false', not {val}")


# Internal

_LOCAL_TECTON_CONFIG_FILE = Path(os.environ.get("TECTON_CONFIG_PATH", Path.home() / ".tecton/config"))
_LOCAL_TECTON_TOKENS_FILE = _LOCAL_TECTON_CONFIG_FILE.with_suffix(".tokens")

_VALID_KEYS_TO_ALLOWED_SOURCES = {
    "API_SERVICE": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_CONFIG,),
    "FEATURE_SERVICE": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_CONFIG,),
    "CLI_CLIENT_ID": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_CONFIG,),
    "TECTON_WORKSPACE": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_CONFIG,),
    "ALPHA_SNOWFLAKE_COMPUTE_ENABLED": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_CONFIG,),
    "TECTON_COMPUTE_MODE": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.REMOTE_MDS_CONFIG,
    ),
    "OAUTH_ACCESS_TOKEN": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_TOKENS,),
    "OAUTH_ACCESS_TOKEN_EXPIRATION": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_TOKENS,),
    "OAUTH_REFRESH_TOKEN": DEFAULT_ALLOWED_SOURCES + (ConfSource.LOCAL_TECTON_TOKENS,),
    # TECTON_CLUSTER_NAME is needed for looking up AWS and Databricks secrets. CLUSTER_REGION is used for looking up AWS
    # secrets. To avoid an infinite loop, these keys cannot be looked up from those sources.
    "TECTON_CLUSTER_NAME": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.REMOTE_MDS_CONFIG,
    ),
    "CLUSTER_REGION": (
        ConfSource.SESSION_OVERRIDE,
        ConfSource.OS_ENV,
        ConfSource.REMOTE_MDS_CONFIG,
        ConfSource.DATABRICKS_SECRET,
    ),
    # NOTE: TECTON_CONF* are meant for Tecton internal use.
    # TODO(TEC-8744): improve tecton.conf configurations such that end users have more
    # control of where secrets are fetched from.
    "SPARK_REDSHIFT_TEMP_DIR": RUNTIME_ALLOWED_SOURCES,
    "TECTON_CONF_DISABLE_DBUTILS": RUNTIME_ALLOWED_SOURCES,
    "TECTON_CONF_DISABLE_AWS_SECRETS": RUNTIME_ALLOWED_SOURCES,
    "TECTON_DEBUG": RUNTIME_ALLOWED_SOURCES,
    "TECTON_RUNTIME_ENV": RUNTIME_ALLOWED_SOURCES,
    "TECTON_RUNTIME_MODE": RUNTIME_ALLOWED_SOURCES,
    "TECTON_VALIDATION_MODE": RUNTIME_ALLOWED_SOURCES,
    "TECTON_FORCE_FUNCTION_SERIALIZATION": RUNTIME_ALLOWED_SOURCES,
    "ALPHA_ATHENA_COMPUTE_ENABLED": DEFAULT_ALLOWED_SOURCES,
    "SQL_DIALECT": DEFAULT_ALLOWED_SOURCES,
    "ATHENA_S3_PATH": DEFAULT_ALLOWED_SOURCES,
    "ATHENA_DATABASE": DEFAULT_ALLOWED_SOURCES,
    "ENABLE_TEMPO": DEFAULT_ALLOWED_SOURCES,
    "QUERY_REWRITE_ENABLED": DEFAULT_ALLOWED_SOURCES,
    "ALPHA_SNOWFLAKE_SNOWPARK_ENABLED": DEFAULT_ALLOWED_SOURCES,
    "AWS_ACCESS_KEY_ID": DEFAULT_ALLOWED_SOURCES,
    "AWS_SECRET_ACCESS_KEY": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_HOST": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_PORT": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_USERNAME": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_DATABASE": DEFAULT_ALLOWED_SOURCES,
    "HIVE_METASTORE_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SPARK_DRIVER_LOCAL_IP": DEFAULT_ALLOWED_SOURCES,
    "METADATA_SERVICE": DEFAULT_ALLOWED_SOURCES,
    "TECTON_API_KEY": DEFAULT_ALLOWED_SOURCES,
    "QUERYTREE_SHORT_SQL_ENABLED": RUNTIME_ALLOWED_SOURCES,
    "REDSHIFT_USER": DEFAULT_ALLOWED_SOURCES,
    "REDSHIFT_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SKIP_FEATURE_TIMESTAMP_VALIDATION": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_ACCOUNT_IDENTIFIER": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_DEBUG": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_SHORT_SQL_ENABLED": DEFAULT_ALLOWED_SOURCES,  # Whether to break up long SQL statements into multiple queries for Snowflake
    "SNOWFLAKE_USER": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_PASSWORD": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_WAREHOUSE": DEFAULT_ALLOWED_SOURCES,
    "SNOWFLAKE_DATABASE": DEFAULT_ALLOWED_SOURCES,
}

_VALID_KEY_PREFIXES = ["SECRET_"]


def _snowpark_enabled():
    ## Try to import snowpark, if it fails, return False
    try:
        import snowflake.snowpark

        return "true"
    except ImportError:
        return "false"


_DEFAULTS = {
    "TECTON_COMPUTE_MODE": (lambda: "spark"),
    "TECTON_WORKSPACE": (lambda: "prod"),
    "FEATURE_SERVICE": (lambda: _get("API_SERVICE")),
    "ALPHA_SNOWFLAKE_SNOWPARK_ENABLED": (_snowpark_enabled),
    "ENABLE_TEMPO": (lambda: "false"),
    "QUERY_REWRITE_ENABLED": (lambda: "true"),
    "SQL_DIALECT": (lambda: "spark"),
    "TECTON_VALIDATION_MODE": (lambda: "explicit"),
}

_REMOTE_MDS_CONFIGS: _ConfigSettings = {}

_is_running_on_databricks_cache = None
_is_running_on_emr_cache = None
TectonEnv = Enum("TectonEnv", "DATABRICKS EMR UNKNOWN")


def _is_running_on_databricks():
    """Whether we're running in Databricks notebook or not."""
    global _is_running_on_databricks_cache
    if _is_running_on_databricks_cache is None:
        main = __import__("__main__")
        filename = os.path.basename(getattr(main, "__file__", ""))
        is_python_shell = filename == "PythonShell.py"
        is_databricks_env = "DBUtils" in main.__dict__
        _is_running_on_databricks_cache = is_python_shell and is_databricks_env
    return _is_running_on_databricks_cache


def _is_running_on_emr():
    """Whether we're running in EMR notebook or not."""
    global _is_running_on_emr_cache
    if _is_running_on_emr_cache is None:
        _is_running_on_emr_cache = "EMR_CLUSTER_ID" in os.environ
    return _is_running_on_emr_cache


def _set_tecton_runtime_env():
    key = "TECTON_RUNTIME_ENV"
    if _is_running_on_databricks():
        set(key, "DATABRICKS")
    elif _is_running_on_emr():
        set(key, "EMR")
    else:
        set(key, "UNKNOWN")


def _is_mode_materialization():
    runtime_mode = get_or_none("TECTON_RUNTIME_MODE")
    return runtime_mode == "MATERIALIZATION"


def _get_runtime_env():
    # Use _get_runtime_only() and not _get() to avoid an infinite loop.
    runtime_env = _get_runtime_only("TECTON_RUNTIME_ENV")
    if runtime_env == "DATABRICKS":
        return TectonEnv.DATABRICKS
    elif runtime_env == "EMR":
        return TectonEnv.EMR
    else:
        return TectonEnv.UNKNOWN


def _get_dbutils():
    # Returns dbutils import. Only works in Databricks notebook environment
    import IPython

    return IPython.get_ipython().user_ns["dbutils"]


def _get_keys_with_allowed_source(source: ConfSource) -> List[str]:
    """Returns all the keys that have the allowed ConfSource."""
    return [key for key, allowed_sources in _VALID_KEYS_TO_ALLOWED_SOURCES.items() if source in allowed_sources]


def save_tecton_configs():
    _save_tecton_config(_LOCAL_TECTON_CONFIG_FILE, _get_keys_with_allowed_source(ConfSource.LOCAL_TECTON_CONFIG))
    _save_tecton_config(_LOCAL_TECTON_TOKENS_FILE, _get_keys_with_allowed_source(ConfSource.LOCAL_TECTON_TOKENS))


def _save_tecton_config(path: Path, keys: Iterable[str]):
    tecton_config = {key: get_or_none(key) for key in keys if get_or_none(key) is not None}
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as f:
        json.dump(tecton_config, f, sort_keys=True, indent=2)
        f.write("\n")


# Get key by looking in TECTON_CLUSTER_NAME'd scope and falling back to "tecton"
def _get_secret_scopes():
    cluster_name = get_or_none("TECTON_CLUSTER_NAME")
    secret_scopes = []
    if cluster_name:
        secret_prefix = cluster_name if cluster_name.startswith("tecton-") else f"tecton-{cluster_name}"
        secret_scopes.append(secret_prefix)
    secret_scopes.append("tecton")
    return secret_scopes


def _get_from_secretsmanager(key: str, scope: str):
    try:
        # Try to Grab secret from AWS secrets manager
        from boto3 import client

        if _is_mode_materialization():
            aws_secret_client = client("secretsmanager")
        else:
            aws_secret_client = client("secretsmanager", region_name=get_or_none("CLUSTER_REGION"))
        secret = aws_secret_client.get_secret_value(SecretId=f"{scope}/{key}")
        return secret["SecretString"]
    except Exception:
        # Do not fail if secret is not found
        return None


def _get_from_db_secrets(key: str, scope: str):
    try:
        dbutils = _get_dbutils()
        return dbutils.secrets.get(scope, key)
    except Exception:
        return None


def save_okta_tokens(access_token, access_token_expiration, refresh_token=None):
    _set("OAUTH_ACCESS_TOKEN", access_token)
    _set("OAUTH_ACCESS_TOKEN_EXPIRATION", access_token_expiration)
    if refresh_token:
        _set("OAUTH_REFRESH_TOKEN", refresh_token)
    _save_tecton_config(_LOCAL_TECTON_TOKENS_FILE, _get_keys_with_allowed_source(ConfSource.LOCAL_TECTON_TOKENS))


def _read_json_config(file_path: Path) -> _ConfigSettings:
    """If the file exists, reads it and returns parsed JSON. Otherwise returns empty dictionary."""
    if not file_path.exists():
        return {}
    content = file_path.read_text()
    if not content:
        return {}
    try:
        return json.loads(content)
    except json.decoder.JSONDecodeError as e:
        raise ValueError(
            f"Unable to decode JSON configuration file {file_path} ({str(e)}). "
            + "To regenerate configuration, delete this file and run `tecton login`."
        )


def validate_api_service_url(url: str):
    """Validate Tecton API URL.
    Returns nothing for valid URLs or raises an error."""
    if "localhost" in url or "ingress" in url:
        return
    if not url.endswith("/api"):
        raise errors.TectonAPIValidationError(
            f'Tecton API URL ("{url}") should be formatted "https://<deployment>.tecton.ai/api"'
        )


# Config values written to and read from the local .tecton/config file.
_LOCAL_TECTON_CONFIG: _ConfigSettings = {}

# Config values read from the local .tecton/config.tokens file.
_LOCAL_TECTON_TOKENS: _ConfigSettings = {}


def _init_configs():
    if not _is_mode_materialization():
        global _LOCAL_TECTON_CONFIG
        _LOCAL_TECTON_CONFIG = _read_json_config(_LOCAL_TECTON_CONFIG_FILE)

        global _LOCAL_TECTON_TOKENS
        _LOCAL_TECTON_TOKENS = _read_json_config(_LOCAL_TECTON_TOKENS_FILE)


def _init_metadata_server_config(mds_response):
    global _REMOTE_MDS_CONFIGS
    _REMOTE_MDS_CONFIGS = dict(mds_response.key_values)


_init_configs()



from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import StrictUndefined

from tecton_core import data_types
from tecton_core.aggregation_utils import get_aggregation_function_name
from tecton_snowflake.snowflake_type_utils import tecton_type_to_snowflake_type


def snowflake_function(value):
    fn = get_aggregation_function_name(value)
    if fn == "mean":
        return "avg"
    return fn


def load_template(name):
    env = Environment(
        loader=PackageLoader("tecton_snowflake"),
        autoescape=False,
        undefined=StrictUndefined,
    )
    env.globals["data_types"] = data_types
    env.filters["snowflake_function"] = snowflake_function
    env.globals["tecton_type_to_snowflake_type"] = tecton_type_to_snowflake_type
    return env.get_template(name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ftecton_proto/auth/service.proto\x12\x11tecton_proto.auth\x1a google/protobuf/descriptor.proto\"\xff\x03\n\x0c\x41uthMetadata\x12\x36\n\x13skip_authentication\x18\x01 \x01(\x08:\x05\x66\x61lseR\x12skipAuthentication\x12\x34\n\x12skip_authorization\x18\x02 \x01(\x08:\x05\x66\x61lseR\x11skipAuthorization\x12.\n\npermission\x18\x05 \x01(\t:\x0enot_configuredR\npermission\x12i\n\x1d\x61\x64vanced_permission_overrides\x18\x06 \x03(\x0b\x32%.tecton_proto.auth.PermissionOverrideR\x1b\x61\x64vancedPermissionOverrides\x12S\n\x12resource_reference\x18\x03 \x01(\x0b\x32$.tecton_proto.auth.ResourceReferenceR\x11resourceReference\x12\x45\n\x1cLEGACY_require_admin_api_key\x18\x07 \x01(\x08:\x05\x66\x61lseR\x18LEGACYRequireAdminApiKey\x12J\n\x1e\x64\x65\x66\x65r_authorization_to_service\x18\x08 \x01(\x08:\x05\x66\x61lseR\x1b\x64\x65\x66\x65rAuthorizationToService\"\xa0\x01\n\x12PermissionOverride\x12\x30\n\x14\x63ondition_field_path\x18\x01 \x01(\tR\x12\x63onditionFieldPath\x12\'\n\x0f\x63ondition_value\x18\x02 \x01(\tR\x0e\x63onditionValue\x12/\n\x13permission_override\x18\x03 \x01(\tR\x12permissionOverride\"c\n\x11ResourceReference\x12:\n\x04type\x18\x01 \x01(\x0e\x32&.tecton_proto.auth.ResourceRefTypeEnumR\x04type\x12\x12\n\x04path\x18\x02 \x01(\tR\x04path*\x84\x01\n\x13ResourceRefTypeEnum\x12\x1d\n\x19RESOURCE_REF_TYPE_UNKNOWN\x10\x00\x12$\n RESOURCE_REF_TYPE_WORKSPACE_NAME\x10\x01\x12(\n$RESOURCE_REF_TYPE_SERVICE_ACCOUNT_ID\x10\x02:g\n\rauth_metadata\x12\x1e.google.protobuf.MethodOptions\x18\xc4\xe7\x8cX \x01(\x0b\x32\x1f.tecton_proto.auth.AuthMetadataR\x0c\x61uthMetadata')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(auth_metadata)

  DESCRIPTOR._options = None
  _RESOURCEREFTYPEENUM._serialized_start=867
  _RESOURCEREFTYPEENUM._serialized_end=999
  _AUTHMETADATA._serialized_start=89
  _AUTHMETADATA._serialized_end=600
  _PERMISSIONOVERRIDE._serialized_start=603
  _PERMISSIONOVERRIDE._serialized_end=763
  _RESOURCEREFERENCE._serialized_start=765
  _RESOURCEREFERENCE._serialized_end=864
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/principal.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!tecton_proto/auth/principal.proto\x12\x11tecton_proto.auth\"d\n\tPrincipal\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12\x0e\n\x02id\x18\x02 \x01(\tR\x02id\"\xdc\x01\n\x0ePrincipalBasic\x12\x32\n\x04user\x18\x03 \x01(\x0b\x32\x1c.tecton_proto.auth.UserBasicH\x00R\x04user\x12Q\n\x0fservice_account\x18\x04 \x01(\x0b\x32&.tecton_proto.auth.ServiceAccountBasicH\x00R\x0eserviceAccount\x12\x35\n\x05group\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.auth.GroupBasicH\x00R\x05groupB\x0c\n\nbasic_info\"\xe9\x01\n\x13ServiceAccountBasic\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12\x1b\n\tis_active\x18\x04 \x01(\x08R\x08isActive\x12\x36\n\x07\x63reator\x18\x05 \x01(\x0b\x32\x1c.tecton_proto.auth.PrincipalR\x07\x63reator\x12\x37\n\x05owner\x18\x06 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\x05owner\"\x81\x01\n\tUserBasic\x12\x17\n\x07okta_id\x18\x01 \x01(\tR\x06oktaId\x12\x1d\n\nfirst_name\x18\x02 \x01(\tR\tfirstName\x12\x1b\n\tlast_name\x18\x03 \x01(\tR\x08lastName\x12\x1f\n\x0blogin_email\x18\x04 \x01(\tR\nloginEmail\"0\n\nGroupBasic\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name*\x82\x01\n\rPrincipalType\x12\x1a\n\x16PRINCIPAL_TYPE_UNKNOWN\x10\x00\x12\x17\n\x13PRINCIPAL_TYPE_USER\x10\x01\x12\"\n\x1ePRINCIPAL_TYPE_SERVICE_ACCOUNT\x10\x02\x12\x18\n\x14PRINCIPAL_TYPE_GROUP\x10\x03\x42;\n\x0f\x63om.tecton.authP\x01Z&github.com/tecton-ai/tecton_proto/auth')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.principal_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authP\001Z&github.com/tecton-ai/tecton_proto/auth'
  _PRINCIPALTYPE._serialized_start=800
  _PRINCIPALTYPE._serialized_end=930
  _PRINCIPAL._serialized_start=56
  _PRINCIPAL._serialized_end=156
  _PRINCIPALBASIC._serialized_start=159
  _PRINCIPALBASIC._serialized_end=379
  _SERVICEACCOUNTBASIC._serialized_start=382
  _SERVICEACCOUNTBASIC._serialized_end=615
  _USERBASIC._serialized_start=618
  _USERBASIC._serialized_end=747
  _GROUPBASIC._serialized_start=749
  _GROUPBASIC._serialized_end=797
# @@protoc_insertion_point(module_scope)



import json
from typing import List

from pyspark.sql.types import DataType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType

from tecton_core.logger import get_logger
from tecton_proto.common.spark_schema_pb2 import SparkSchema

logger = get_logger("SparkSchemaWrapper")


class SparkSchemaWrapper:
    """
    Wrapper around Spark schema (StructType).
    """

    def __init__(self, schema: StructType):
        """
        Creates a new SparkSchemaWrapper.

        :param schema: DataFrame schema.
        """
        self._schema = schema

    def to_proto(self) -> SparkSchema:
        return self.from_spark_schema(self._schema)

    @classmethod
    def from_spark_schema(cls, schema) -> SparkSchema:
        proto = SparkSchema()
        spark_dict = schema.jsonValue()
        for field in spark_dict["fields"]:
            proto_field = proto.fields.add()
            proto_field.name = field["name"]
            proto_field.structfield_json = json.dumps(field)
        return proto

    @classmethod
    def from_proto(cls, proto: SparkSchema) -> "SparkSchemaWrapper":
        s = StructType()
        for field in proto.fields:
            field_schema = StructField.fromJson(json.loads(field.structfield_json))
            s.add(field_schema)
        return SparkSchemaWrapper(s)

    def unwrap(self) -> StructType:
        return self._schema

    def column_names(self) -> List[str]:
        return [f["name"] for f in self._schema.jsonValue()["fields"]]

    def spark_type(self, column: str) -> DataType:
        return self._schema[column].dataType

    def column_name_types(self):
        return [(c, self.spark_type(c)) for c in self.column_names()]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/user.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ctecton_proto/data/user.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\"\xae\x02\n\x04User\x12\x17\n\x07okta_id\x18\x01 \x01(\tR\x06oktaId\x12\x1d\n\nfirst_name\x18\x02 \x01(\tR\tfirstName\x12\x1b\n\tlast_name\x18\x03 \x01(\tR\x08lastName\x12\x1f\n\x0blogin_email\x18\x04 \x01(\tR\nloginEmail\x12\x1f\n\x0bokta_status\x18\x05 \x01(\tR\noktaStatus\x12\x39\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x19\n\x08is_admin\x18\x07 \x01(\x08R\x07isAdmin\x12\x39\n\nlast_login\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tlastLoginB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.user_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _USER._serialized_start=85
  _USER._serialized_end=387
# @@protoc_insertion_point(module_scope)
